@@ -33,7 +33,7 @@ except ImportError:
 
 __all__ = [
     'ImproperlyConfigured', 'SqliteDatabase', 'MySQLDatabase', 'PostgresqlDatabase',
-    'asc', 'desc', 'Count', 'Max', 'Min', 'Q', 'Field', 'CharField', 'TextField',
+    'asc', 'desc', 'Count', 'Max', 'Min', 'Sum', 'Q', 'Field', 'CharField', 'TextField',
     'DateTimeField', 'BooleanField', 'DecimalField', 'FloatField', 'IntegerField',
     'PrimaryKeyField', 'ForeignKeyField', 'Model', 'filter_query', 'annotate_query',
 ]
@@ -434,6 +434,9 @@ def Max(f, alias='max'):
 def Min(f, alias='min'):
     return ('MIN', f, alias)
 
+def Sum(f, alias='sum'):
+    return ('SUM', f, alias)
+
 # decorator for query methods to indicate that they change the state of the
 # underlying data structures
 def returns_clone(func):
@@ -448,9 +451,9 @@ ternary = lambda cond, t, f: (cond and [t] or [f])[0]
 
 
 class Node(object):
-    def __init__(self, connector='AND'):
+    def __init__(self, connector='AND', children=None):
         self.connector = connector
-        self.children = []
+        self.children = children or []
         self.negated = False
     
     def connect(self, rhs, connector):
@@ -476,6 +479,9 @@ class Node(object):
     def __invert__(self):
         self.negated = not self.negated
         return self
+
+    def __nonzero__(self):
+        return bool(self.children)
     
     def __unicode__(self):
         query = []
@@ -494,7 +500,8 @@ class Node(object):
     
 
 class Q(object):
-    def __init__(self, **kwargs):
+    def __init__(self, _model=None, **kwargs):
+        self.model = _model
         self.query = kwargs
         self.parent = None
         self.negated = False
@@ -528,20 +535,38 @@ class Q(object):
         return expr
 
 
-def parseq(*args, **kwargs):
+def apply_model(model, item):
+    if isinstance(item, Node):
+        for child in item.children:
+            apply_model(model, child)
+    elif isinstance(item, Q):
+        if item.model is None:
+            item.model = model
+
+def parseq(model, *args, **kwargs):
     node = Node()
     
     for piece in args:
+        apply_model(model, piece)
         if isinstance(piece, (Q, Node)):
             node.children.append(piece)
         else:
             raise TypeError('Unknown object: %s', piece)
 
     if kwargs:
-        node.children.append(Q(**kwargs))
+        node.children.append(Q(model, **kwargs))
 
     return node
 
+def find_models(item):
+    seen = set()
+    if isinstance(item, Node):
+        for child in item.children:
+            seen.update(find_models(child))
+    elif isinstance(item, Q):
+        seen.add(item.model)
+    return seen
+
 
 class EmptyResultException(Exception):
     pass
@@ -560,7 +585,8 @@ class BaseQuery(object):
         self.interpolation = self.database.adapter.interpolation
         
         self._dirty = True
-        self._where = {}
+        self._where = []
+        self._where_models = set()
         self._joins = {}
         self._joined_models = set()
     
@@ -571,7 +597,7 @@ class BaseQuery(object):
         return cloned
     
     def clone_where(self):
-        return self._clone_dict_graph(self._where)
+        return list(self._where)
     
     def clone_joins(self):
         return self._clone_dict_graph(self._joins)
@@ -624,8 +650,10 @@ class BaseQuery(object):
     
     @returns_clone
     def where(self, *args, **kwargs):
-        self._where.setdefault(self.query_context, [])
-        self._where[self.query_context].append(parseq(*args, **kwargs))
+        parsed = parseq(self.query_context, *args, **kwargs)
+        if parsed:
+            self._where.append(parsed)
+            self._where_models.update(find_models(parsed))
 
     @returns_clone
     def join(self, model, join_type=None, on=None):
@@ -684,7 +712,7 @@ class BaseQuery(object):
                 right_field = field.name
             
             if join_type is None:
-                if field.null and model not in self._where:
+                if field.null and model not in self._where_models:
                     join_type = 'LEFT OUTER'
                 else:
                     join_type = 'INNER'
@@ -708,9 +736,6 @@ class BaseQuery(object):
         alias_map = {}
 
         alias_required = self.use_aliases()
-
-        where_with_alias = []
-        where_data = []
         
         if alias_required:
             alias_count += 1
@@ -720,13 +745,17 @@ class BaseQuery(object):
         
         computed_joins = self.follow_joins(self.model, alias_map, alias_required, alias_count)
         
-        for model in sorted(self._where, key=lambda m: alias_map[m]):
-            for node in self._where[model]:
-                query, data = self.parse_node(node, model, alias_map)
-                where_with_alias.append(query)
-                where_data.extend(data)
+        clauses = [self.parse_node(node, alias_map) for node in self._where]
         
-        return computed_joins, where_with_alias, where_data, alias_map
+        return computed_joins, clauses, alias_map
+    
+    def flatten_clauses(self, clauses):
+        where_with_alias = []
+        where_data = []
+        for query, data in clauses:
+            where_with_alias.append(query)
+            where_data.extend(data)
+        return where_with_alias, where_data
     
     def convert_where_to_params(self, where_data):
         flattened = []
@@ -737,17 +766,17 @@ class BaseQuery(object):
                 flattened.append(clause)
         return flattened
     
-    def parse_node(self, node, model, alias_map):
+    def parse_node(self, node, alias_map):
         query = []
         query_data = []
         nodes = []
         for child in node.children:
             if isinstance(child, Q):
-                parsed, data = self.parse_q(child, model, alias_map)
+                parsed, data = self.parse_q(child, alias_map)
                 query.append(parsed)
                 query_data.extend(data)
             elif isinstance(child, Node):
-                parsed, data = self.parse_node(child, model, alias_map)
+                parsed, data = self.parse_node(child, alias_map)
                 query.append('(%s)' % parsed)
                 query_data.extend(data)
         query.extend(nodes)
@@ -757,7 +786,8 @@ class BaseQuery(object):
             query = 'NOT (%s)' % query
         return query, query_data
     
-    def parse_q(self, q, model, alias_map):
+    def parse_q(self, q, alias_map):
+        model = q.model or self.model
         query = []
         query_data = []
         parsed = self.parse_query_args(model, **q.query)
@@ -828,7 +858,8 @@ class SelectQuery(BaseQuery):
         self._group_by = []
         self._having = []
         self._order_by = []
-        self._pagination = None # return all by default
+        self._limit = None
+        self._offset = None
         self._distinct = False
         self._qr = None
         super(SelectQuery, self).__init__(model)
@@ -839,21 +870,34 @@ class SelectQuery(BaseQuery):
         query._group_by = list(self._group_by)
         query._having = list(self._having)
         query._order_by = list(self._order_by)
-        query._pagination = self._pagination and tuple(self._pagination) or None
+        query._limit = self._limit
+        query._offset = self._offset
         query._distinct = self._distinct
         query._qr = self._qr
         query._where = self.clone_where()
+        query._where_models = set(self._where_models)
         query._joined_models = self._joined_models.copy()
         query._joins = self.clone_joins()
         return query
     
     @returns_clone
-    def paginate(self, page_num, paginate_by=20):
-        self._pagination = (page_num, paginate_by)
+    def paginate(self, page, paginate_by=20):
+        if page > 0:
+            page -= 1
+        self._limit = paginate_by
+        self._offset = page * paginate_by
+    
+    @returns_clone
+    def limit(self, num_rows):
+        self._limit = num_rows
+    
+    @returns_clone
+    def offset(self, num_rows):
+        self._offset = num_rows
     
     def count(self):
-        tmp_pagination = self._pagination
-        self._pagination = None
+        tmp_lim, tmp_off = self._limit, self._offset
+        self._limit = self._offset = None
         
         tmp_query = self.query
         
@@ -864,8 +908,10 @@ class SelectQuery(BaseQuery):
         
         res = self.database.execute(*self.sql())
         
+        # restore
         self.query = tmp_query
-        self._pagination = tmp_pagination
+        self._limit = tmp_lim
+        self._offset = tmp_off
         
         return res.fetchone()[0]
     
@@ -972,7 +1018,8 @@ class SelectQuery(BaseQuery):
             raise TypeError('Unknown type encountered parsing select query')
     
     def sql(self):
-        joins, where, where_data, alias_map = self.compile_where()
+        joins, clauses, alias_map = self.compile_where()
+        where, where_data = self.flatten_clauses(clauses)
         
         table = self.model._meta.db_table
 
@@ -1022,11 +1069,10 @@ class SelectQuery(BaseQuery):
             pieces.append('HAVING %s' % having)
         if order_by:
             pieces.append('ORDER BY %s' % ', '.join(order_by))
-        if self._pagination:
-            page, paginate_by = self._pagination
-            if page > 0:
-                page -= 1
-            pieces.append('LIMIT %d OFFSET %d' % (paginate_by, page * paginate_by))
+        if self._limit:
+            pieces.append('LIMIT %d' % self._limit)
+        if self._offset:
+            pieces.append('OFFSET %d' % self._offset)
         
         return ' '.join(pieces), params
     
@@ -1054,6 +1100,7 @@ class UpdateQuery(BaseQuery):
     def clone(self):
         query = UpdateQuery(self.model, **self.update_query)
         query._where = self.clone_where()
+        query._where_models = set(self._where_models)
         query._joined_models = self._joined_models.copy()
         query._joins = self.clone_joins()
         return query
@@ -1073,7 +1120,8 @@ class UpdateQuery(BaseQuery):
         return sets
     
     def sql(self):
-        joins, where, where_data, alias_map = self.compile_where()
+        joins, clauses, alias_map = self.compile_where()
+        where, where_data = self.flatten_clauses(clauses)
         set_statement = self.parse_update()
 
         params = []
@@ -1107,12 +1155,14 @@ class DeleteQuery(BaseQuery):
     def clone(self):
         query = DeleteQuery(self.model)
         query._where = self.clone_where()
+        query._where_models = set(self._where_models)
         query._joined_models = self._joined_models.copy()
         query._joins = self.clone_joins()
         return query
     
     def sql(self):
-        joins, where, where_data, alias_map = self.compile_where()
+        joins, clauses, alias_map = self.compile_where()
+        where, where_data = self.flatten_clauses(clauses)
 
         params = []
         
@@ -1194,8 +1244,6 @@ def convert_lookup(model, joins, lookup):
         
         # we have some joins
         if len(pieces):
-            query_model = model
-            
             for piece in pieces:
                 # piece is something like 'blog' or 'entry_set'
                 joined_model = None
@@ -1234,26 +1282,23 @@ def filter_query(model_or_query, *args, **kwargs):
     query = {} # mapping of models to queries
     joins = {} # a graph of joins needed, passed into the convert_lookup function
     
-    # due to quirks in the way where clauses are defined, Q() queries can only
-    # work on a single table -- this will need to be redesigned so Q() objects
-    # can work on multiple models
+    # traverse Q() objects, find any joins that may be lurking -- clean up the
+    # lookups and assign the correct model
     def fix_q(node_or_q, joins):
         if isinstance(node_or_q, Node):
             for child in node_or_q.children:
-                query_model = fix_q(child, joins)
+                fix_q(child, joins)
         elif isinstance(node_or_q, Q):
             new_query = {}
+            curr_model = node_or_q.model or model
             for raw_lookup, value in node_or_q.query.items():
-                # do we have a query model at this point?
-                query_model, joins, lookup = convert_lookup(model, joins, raw_lookup)
+                query_model, joins, lookup = convert_lookup(curr_model, joins, raw_lookup)
                 new_query[lookup] = value
+            node_or_q.model = query_model
             node_or_q.query = new_query
-        return query_model
     
     for node_or_q in args:
-        queried_model = fix_q(node_or_q, joins)
-        query.setdefault(queried_model, [])
-        query[queried_model].append(node_or_q)
+        fix_q(node_or_q, joins)
     
     # iterate over keyword lookups and determine lookups and necessary joins
     for raw_lookup, value in kwargs.items():
@@ -1271,6 +1316,9 @@ def filter_query(model_or_query, *args, **kwargs):
         return query
     select_query = follow_joins(model, select_query)
     
+    for node in args:
+        select_query = select_query.where(node)
+    
     for model, lookups in query.items():
         qargs, qkwargs = [], {}
         for lookup in lookups:
@@ -1371,6 +1419,9 @@ class Field(object):
     def lookup_value(self, lookup_type, value):
         return self.db_value(value)
 
+    def class_prepared(self):
+        pass
+
 
 class CharField(Field):
     db_field = 'string'
@@ -1479,17 +1530,22 @@ class PrimaryKeyField(IntegerField):
 
 
 class ForeignRelatedObject(object):    
-    def __init__(self, to, name):
-        self.field_name = name
+    def __init__(self, to, field):
         self.to = to
-        self.cache_name = '_cache_%s' % name
+        self.field = field
+        self.field_name = self.field.name
+        self.cache_name = '_cache_%s' % self.field_name
     
     def __get__(self, instance, instance_type=None):
         if not getattr(instance, self.cache_name, None):
             id = getattr(instance, self.field_name, 0)
-            qr = self.to.select().where(**{self.to._meta.pk_name: id}).execute()
-            setattr(instance, self.cache_name, qr.next())
-        return getattr(instance, self.cache_name)
+            qr = self.to.select().where(**{self.to._meta.pk_name: id})
+            try:
+                setattr(instance, self.cache_name, qr.get())
+            except self.to.DoesNotExist:
+                if not self.field.null:
+                    raise
+        return getattr(instance, self.cache_name, None)
     
     def __set__(self, instance, obj):
         assert isinstance(obj, self.to), "Cannot assign %s, invalid type" % obj
@@ -1510,14 +1566,17 @@ class ReverseForeignRelatedObject(object):
 
 class ForeignKeyField(IntegerField):
     db_field = 'foreign_key'
-    field_template = '%(column_type)s%(nullable)s REFERENCES %(to_table)s (%(to_pk)s)'
+    field_template = '%(column_type)s%(nullable)s REFERENCES %(to_table)s (%(to_pk)s)%(cascade)s%(extra)s'
     
-    def __init__(self, to, null=False, related_name=None, *args, **kwargs):
+    def __init__(self, to, null=False, related_name=None, cascade=False, extra=None, *args, **kwargs):
         self.to = to
         self.related_name = related_name
+        self.cascade = cascade
+        self.extra = extra
+
         kwargs.update({
-            'to_table': to._meta.db_table,
-            'to_pk': to._meta.pk_name
+            'cascade': ' ON DELETE CASCADE' if self.cascade else '',
+            'extra': self.extra or '',
         })
         super(ForeignKeyField, self).__init__(null=null, *args, **kwargs)
     
@@ -1525,14 +1584,17 @@ class ForeignKeyField(IntegerField):
         self.descriptor = name
         self.name = name + '_id'
         self.model = klass
-        
+
+        if self.to == 'self':
+            self.to = self.model
+
         self.verbose_name = self.verbose_name or re.sub('_', ' ', name).title()
         
         if self.related_name is None:
             self.related_name = klass._meta.db_table + '_set'
         
         klass._meta.rel_fields[name] = self.name
-        setattr(klass, self.descriptor, ForeignRelatedObject(self.to, self.name))
+        setattr(klass, self.descriptor, ForeignRelatedObject(self.to, self))
         setattr(klass, self.name, None)
         
         reverse_rel = ReverseForeignRelatedObject(klass, self.name)
@@ -1549,12 +1611,23 @@ class ForeignKeyField(IntegerField):
             return value.get_pk()
         return value
 
+    def class_prepared(self):
+        # unfortunately because we may not know the primary key field
+        # at the time this field's add_to_class() method is called, we
+        # need to update the attributes after the class has been built
+        self.attributes.update({
+            'to_table': self.to._meta.db_table,
+            'to_pk': self.to._meta.pk_name,
+        })
+
 
 # define a default database object in the module scope
 database = SqliteDatabase(DATABASE_NAME)
 
 
 class BaseModelOptions(object):
+    ordering = None
+
     def __init__(self, model_class, options=None):
         # configurable options
         options = options or {'database': database}
@@ -1602,7 +1675,7 @@ class BaseModelOptions(object):
 
 
 class BaseModel(type):
-    inheritable_options = ['database']
+    inheritable_options = ['database', 'ordering']
     
     def __new__(cls, name, bases, attrs):
         cls = super(BaseModel, cls).__new__(cls, name, bases, attrs)
@@ -1644,6 +1717,9 @@ class BaseModel(type):
             _meta.fields[_meta.pk_name] = pk
 
         _meta.model_name = cls.__name__
+
+        for field in _meta.fields.values():
+            field.class_prepared()
                 
         if hasattr(cls, '__unicode__'):
             setattr(cls, '__repr__', lambda self: '<%s: %s>' % (
@@ -1659,6 +1735,7 @@ class Model(object):
     __metaclass__ = BaseModel
     
     def __init__(self, *args, **kwargs):
+        self.get_field_dict()
         for k, v in kwargs.items():
             setattr(self, k, v)
     
@@ -1710,7 +1787,10 @@ class Model(object):
     
     @classmethod
     def select(cls, query=None):
-        return SelectQuery(cls, query)
+        select_query = SelectQuery(cls, query)
+        if cls._meta.ordering:
+            select_query = select_query.order_by(*cls._meta.ordering)
+        return select_query
     
     @classmethod
     def update(cls, **query):
